#include "config_file.h"

char *Colors[]={"none","black","red","green","yellow","blue","magenta","cyan","white",NULL};

int LoadColor(char *ColorString)
{
char *ptr;
int val=0;

	ptr=strchr(ColorString,'/');
	if (ptr)
	{
		*ptr='\0';
		ptr++;
		val=MatchTokenFromList(ptr,Colors,0);
		val=val << 8;
	}
	val |=MatchTokenFromList(ColorString,Colors,0);

return(val);
}



TAction *NewCrayonAction(TCrayon *Crayon)
{
if ((Crayon->ActionCount % 10)==0) 
{
	Crayon->Actions=(TAction *) calloc(Crayon->ActionCount+10, sizeof(TAction));
}

Crayon->ActionCount++;

return(Crayon->Actions + Crayon->ActionCount-1);
}


//This parses those actions that take a string
int ParseStringAction(int Type, char *EnvName, TCrayon *Crayon)
{
	TAction *Action;

		Action=NewCrayonAction(Crayon);
		Action->Attribs |= Type;
		if (Type==FLAG_SETENV) Action->Op='e';
		if (Type==FLAG_SETSTR) Action->Op='s';
		Action->String=CopyStr(Action->String,EnvName);
		return(Type);
}



TAction *ParseAttribs(TCrayon *Crayon, TAction *Action, char *Operations)
{
char *Token=NULL, *ptr;
int Attribs=0;
int val=0;
 
//Do this before anything else!

ptr=GetToken(Operations," ",&Token,0);
while (ptr)
{		
	if (strcasecmp(Token,"caps")==0) Action->Attribs |= FLAG_CAPS;
	else if (strcasecmp(Token,"bold")==0) Action->Attribs |= FLAG_BOLD;
	else if (strcasecmp(Token,"hide")==0) Action->Attribs |= FLAG_HIDE;
	else if (strcasecmp(Token,"blink")==0) Action->Attribs |= FLAG_BLINK;
	else if (strcasecmp(Token,"uppercase")==0) Action->Attribs |= FLAG_CAPS;
	else if (strcasecmp(Token,"underline")==0) Action->Attribs |= FLAG_UNDER;
	else if (strcasecmp(Token,"lowercase")==0) Action->Attribs |= FLAG_LOWERCASE;
	else if (strcasecmp(Token,"inverse")==0) Action->Attribs |= FLAG_INVERSE;
	else if (strcasecmp(Token,"setxtitle")==0) Action->Flags |= FLAG_SET_XTITLE;
	else if (strcasecmp(Token,"basename")==0) 
	{
		Action=NewCrayonAction(Crayon);
		Action->Flags |= FLAG_BASENAME;
	}
	else if (strcasecmp(Token,"restorextitle")==0) 
	{
		Action->Flags |= FLAG_RESTORE_XTITLE;
		GlobalFlags |= FLAG_RESTORE_XTITLE;
	}
	else if (strcasecmp(Token,"setstr")==0) 
	{
		ptr=GetToken(ptr," ",&Token,0);
		Action->Flags |= ParseStringAction(FLAG_SETSTR,Token,Crayon);
	}
	else if (strcasecmp(Token,"setenv")==0) 
	{
		ptr=GetToken(ptr," ",&Token,0);
		Action->Flags |= ParseStringAction(FLAG_SETENV,Token,Crayon);
	}
	else Action->Attribs |= LoadColor(Token);
ptr=GetToken(ptr," ",&Token,0);
}

DestroyString(Token);

return(Action);
}



// Token will either be a test (with an operation like '<' '=' etc) or a list of
// actions (colors, uppercase) to apply
void ParseActionToken(char *Operations, TCrayon *Crayon)
{
char *Token=NULL, *ptr;
TAction *Action=NULL;
int val=0;
 
//Do this before anything else!

ptr=GetToken(Operations," ",&Token,0);
Action=NewCrayonAction(Crayon);
while (ptr)
{		
		switch (*Token)
		{
			case '=':
			case '!':
			case '<':
			case '>':
				if (Action->Op != '\0') Action=NewCrayonAction(Crayon);
				Action->Op=*Token;
				Action->Value=atof(Token+1);
			break;

			default:
				Action=ParseAttribs(Crayon,Action,Token);
			break;
		}
ptr=GetToken(ptr," ",&Token,0);
}

DestroyString(Token);
}



void ParseCrayonAction(TCrayon *Item, char *Args)
{
char *Token=NULL, *ptr;

ptr=GetToken(Args,"\\S",&Item->Match,GETTOKEN_QUOTES);

switch (Item->Type)
{
	case CRAYON_LINE:
	case CRAYON_STRING:
	case CRAYON_APPEND:
	case CRAYON_PREPEND:
	case CRAYON_VALUE:
	case CRAYON_EXEC:
	case CRAYON_CMDLINE:
		ParseActionToken(ptr, Item);
	break;


	case CRAYON_MAPTO:
	case CRAYON_LINEMAPTO:
	ptr=GetToken(ptr,"\\S",&Token,GETTOKEN_QUOTES);
	while (ptr)
	{
		ParseActionToken(Token, Item);
		ptr=GetToken(ptr,"\\S",&Token,GETTOKEN_QUOTES);
	}
	break;
}

DestroyString(Token);

return(Item);
}


TCrayon *ParseCrayonEntry(char *Token, char *Args, char **UserOptions, ListNode *ColorMatches)
{
	TCrayon *Crayon=NULL;
	int Type=0;

	if (UserOptions && (strcmp(Token,"args")==0)) *UserOptions=CopyStr(*UserOptions,Args);
	else if (strcasecmp(Token,"line")==0) Type=CRAYON_LINE;
	else if (strcasecmp(Token,"string")==0) Type=CRAYON_STRING;
	else if (strcasecmp(Token,"value")==0) Type=CRAYON_VALUE;
	else if (strcasecmp(Token,"mapto")==0) Type=CRAYON_MAPTO;
	else if (strcasecmp(Token,"linemapto")==0) Type=CRAYON_LINEMAPTO;
	else if (strcasecmp(Token,"append")==0) Type=CRAYON_APPEND;
	else if (strcasecmp(Token,"prepend")==0) Type=CRAYON_PREPEND;
	else if (strcasecmp(Token,"exec")==0) Type=CRAYON_EXEC;
	else if (strcasecmp(Token,"cmdline")==0) Type=CRAYON_CMDLINE;

		
	if (Type > 0)
	{
	Crayon=(TCrayon *) calloc(1,sizeof(TCrayon));
	Crayon->Type=Type;
	ParseCrayonAction(Crayon, Args);
	if (ColorMatches) ListAddNamedItem(ColorMatches,Crayon->Match,Crayon);
	}

return(Crayon);
}



void ConfigReadEntry(STREAM *S, char **UserOptions, ListNode *ColorMatches)
{
char *Tempstr=NULL, *Token=NULL, *Arg=NULL, *ptr;
ListNode *Curr;

	Tempstr=STREAMReadLine(Tempstr,S);
	while (Tempstr)
	{
		StripTrailingWhitespace(Tempstr);
		StripLeadingWhitespace(Tempstr);
		ptr=GetToken(Tempstr,"\\S",&Token,0);

		if (strcmp(Token,"}")==0) break;
		ParseCrayonEntry(Token,ptr,UserOptions,ColorMatches);
	Tempstr=STREAMReadLine(Tempstr,S);
	}

DestroyString(Tempstr);
DestroyString(Token);
}


int EntryMatchesCommand(char *EntryList, char *EntryArgs, char *ProgPath, char *ProgArgs)
{
char *Entry=NULL, *ptr, *p_ProgName;
int result=FALSE;

p_ProgName=strrchr(ProgPath,'/');
if (p_ProgName) p_ProgName++;
else p_ProgName=ProgPath;
ptr=GetToken(EntryList,"|",&Entry,0);
while (ptr)
{
	if (strcmp(Entry,p_ProgName)==0)	
	{
		if (StrLen(EntryArgs)==0) 
		{
			result=TRUE;
			break;
		}

		if (pmatch(EntryArgs,ProgArgs,StrLen(ProgArgs),NULL,0) > 0) 
		{
			result=TRUE;
			break;
		}
	}
	ptr=GetToken(ptr,"|",&Entry,0);
}

DestroyString(Entry);
return(result);
}



int ConfigReadFile(char *Path, char *CommandLine, char **UserOptions, char **CrayonizerDir, ListNode *ColorMatches)
{
STREAM *S;
char *Tempstr=NULL, *Token=NULL, *ptr;
char *ProgName=NULL, *Args;


S=STREAMOpenFile(Path,O_RDONLY);
if (! S) return(FALSE);

Args=GetToken(CommandLine," ",&ProgName,0);
Tempstr=STREAMReadLine(Tempstr,S);
while (Tempstr)
{
	StripTrailingWhitespace(Tempstr);
	StripLeadingWhitespace(Tempstr);
	ptr=GetToken(Tempstr,"\\S",&Token,0);
	if (strcasecmp(Token,"CrayonizerDir")==0) *CrayonizerDir=CopyStr(*CrayonizerDir,ptr);
	if (strcasecmp(Token,"entry")==0)
	{
		ptr=GetToken(ptr,"\\S",&Token,0);
		if (EntryMatchesCommand(Token,ptr,ProgName, Args))
		{
			ConfigReadEntry(S,UserOptions,ColorMatches);
		}
		else ConfigReadEntry(S,NULL,NULL);

	}
	if (strcasecmp(Token,"include")==0) ConfigReadFile(ptr, CommandLine, UserOptions, CrayonizerDir, ColorMatches);

	Tempstr=STREAMReadLine(Tempstr,S);
}

DestroyString(ProgName);
DestroyString(Tempstr);
DestroyString(Token);

return(TRUE);
}



