//Crayonizer. A coloration/formatting app for command-line output
//Written by Colum Paget.
//Copyright 2013 Colum Paget.


/****  Gnu Public Licence ****/
/*
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "common.h"
#include "config_file.h"
#include "help.h"


#define NORM "\x1b[0m"
#define CLRSCR "\x1b[2J\x1b[;H"


char *Version="0.0.4";
int GlobalFlags=0;
ListNode *Vars=NULL;

char *VtCode(int Color, int BgColor, int Flags)
{
static char *ANSI=NULL;

if ((! Color) && (! Flags)) return("");

if ( (Color > 0) && (BgColor > 0) )
{	
	BgColor=BgColor >> 8;
	ANSI=FormatStr(ANSI,"\x1b[%d;%d",30+Color-1,40+BgColor-1);
	if (Flags) ANSI=CatStr(ANSI,";");
}
else if (Color > 0) 
{
	ANSI=FormatStr(ANSI,"\x1b[%d",30+Color-1);
	if (Flags) ANSI=CatStr(ANSI,";");
}
else ANSI=CopyStr(ANSI,"\x1b[");

if (Flags)
{
	if (Flags & FLAG_BOLD) ANSI=CatStr(ANSI,"01");
	if (Flags & FLAG_FAINT) ANSI=CatStr(ANSI,"02");
	if (Flags & FLAG_UNDER) ANSI=CatStr(ANSI,"04");
	if (Flags & FLAG_BLINK) ANSI=CatStr(ANSI,"05");
	if (Flags & FLAG_INVERSE) ANSI=CatStr(ANSI,"07");
}
ANSI=CatStr(ANSI,"m");

return(ANSI);
}


char *RebuildPath(char *RetStr, char *Path, char *CurrDir)
{
char *NewPath=NULL, *Token=NULL, *ptr;

NewPath=CopyStr(RetStr,"");
ptr=GetToken(Path,":",&Token,0);
while (ptr)
{
	if (strcmp(Token,CurrDir) !=0) NewPath=MCatStr(NewPath,Token,":",NULL);
	ptr=GetToken(ptr,":",&Token,0);
}

DestroyString(Token);
return(NewPath);
}


char *RebuildCommandLine(char *RetStr,char *UserOptions, int argc, char *argv[])
{
char *CommandLine=NULL;
int i;

CommandLine=MCopyStr(RetStr, argv[0]," ",UserOptions," ",NULL);
for (i=1; i < argc; i++)
{
CommandLine=MCatStr(CommandLine,"'",argv[i],"' ",NULL);
}

return(CommandLine);
}


void LoadEnvironment()
{
char *Token=NULL, *ptr;
int i;

Vars=ListCreate();
for (i=0; environ[i] !=NULL; i++)
{
ptr=GetToken(environ[i],"=",&Token,0);
SetVar(Vars,Token,ptr);
}

DestroyString(Token);
}



int CrayonMatches(TCrayon *Crayon, char *sptr, char *eptr)
{
char *ptr;
int i, result=FALSE;
float val;

if (Crayon->Type==CRAYON_STRING) return(TRUE);
if (Crayon->Type==CRAYON_LINE) return(TRUE);
if (Crayon->Type==CRAYON_MAPTO) return(TRUE);
if (Crayon->Type==CRAYON_LINEMAPTO) return(TRUE);
if (Crayon->Type==CRAYON_CMDLINE) return(TRUE);

val=atof(sptr);

for (i=0; i < Crayon->ActionCount; i++)
{
	switch (Crayon->Actions[i].Op)
	{
	case '<':
	if (val < Crayon->Actions[i].Value) result=TRUE;
	break;
	
	case '>':
	if (val > Crayon->Actions[i].Value) result=TRUE;
	
	break;
	
	case '=':
	if (val == Crayon->Actions[i].Value) result=TRUE;
	break;
	
	case '!':
	if (val != Crayon->Actions[i].Value) result=TRUE;
	break;
	}
}

return(result);
}



void ApplySingleAction(int *AttribLine, char *Line, char *MatchStart, char *MatchEnd, TAction *Action)
{
int Attribs, Flags;
int start, end, i;
char *Tempstr=NULL, *ptr;


	Flags=Action->Flags;
	Attribs=Action->Attribs;
	start=MatchStart-Line;
	end=MatchEnd-Line;

		//These have to be done when we have the particular match in hand, they 
		//can't be done when we output the line at the end
		if (Flags & FLAG_BASENAME) 
		{
			ptr=strrchr(MatchStart,'/');
			if (ptr && (ptr < MatchEnd))
			{
				MatchStart=ptr+1;
				start=MatchStart-Line;
			}
		}

		if (Flags & FLAG_SET_XTITLE) 
		{
			Tempstr=CopyStr(Tempstr,"\x1b]0;");
			Tempstr=CatStrLen(Tempstr,MatchStart,end-start);
			StripTrailingWhitespace(Tempstr);
			Tempstr=CatStr(Tempstr,"\x07");
			write(1,Tempstr,StrLen(Tempstr));
		}

		if (Flags & FLAG_RESTORE_XTITLE) 
		{
			Tempstr=CopyStr(Tempstr,"\x1b]0;");
			Tempstr=CatStr(Tempstr,GetVar(Vars,"OldXtermTitle"));
			StripTrailingWhitespace(Tempstr);
			Tempstr=CatStr(Tempstr,"\x07");
			write(1,Tempstr,StrLen(Tempstr));
		}

		if (Flags & FLAG_SETENV) 
		{
			Tempstr=CopyStrLen(Tempstr,MatchStart,end-start);
			StripTrailingWhitespace(Tempstr);
				//'e' means 'setenv'
				if (Action->Op=='e')
				{
					setenv(Action->String,Tempstr,TRUE);
					SetVar(Vars,Action->String,Tempstr);
				}
		}

		if (AttribLine) for (i=start; i < end; i++) AttribLine[i]|=Attribs;

	DestroyString(Tempstr);
}


void ApplyActions(int *AttribLine, char *Line, char *MatchStart, char *MatchEnd, TCrayon *Crayon)
{
int i,sum=0;
char *ptr;

	//for map to we must calculate which attribute to use
	if ((Crayon->Type==CRAYON_MAPTO) || (Crayon->Type==CRAYON_LINEMAPTO))
	{
		for (ptr=MatchStart; ptr < MatchEnd; ptr++) sum+=*ptr;
		ApplySingleAction(AttribLine, Line, MatchStart, MatchEnd, &Crayon->Actions[sum % Crayon->ActionCount]);
	}
	else 
	{
		for (i=0; i < Crayon->ActionCount; i++)
		{
			ApplySingleAction(AttribLine, Line, MatchStart, MatchEnd, &Crayon->Actions[i]);
		}
	}

}


//AttribLine is the string of attribute values for each char in Line
void ColorSubstring(int *AttribLine, char *Line, int Len, TCrayon *Crayon)
{
char *Tempstr=NULL, *p_LineEnd;
int result;
ListNode *Matches, *Curr;
TPMatch *Match;
int Attribs;

p_LineEnd=Line+Len;

Matches=ListCreate();
result=pmatch(Crayon->Match, Line, p_LineEnd-Line, Matches, PMATCH_SUBSTR);
if (result > 0)
{
	Curr=ListGetNext(Matches);
	while (Curr)
	{
	Match=(TPMatch *) Curr->Item;
	if (CrayonMatches(Crayon, Match->Start,Match->End)) ApplyActions(AttribLine, Line, Match->Start, Match->End, Crayon);
	Curr=ListGetNext(Curr);
	}
}

ListDestroy(Matches,DestroyString);
}


void OutputLineWithAttributes(char *Line, int *Attribs, int Len)
{
int i, LastAttrib=0;
char *Tempstr=NULL;

if (! Attribs) return;
if (Attribs[0] & FLAG_HIDE) return;

for (i=0; i < Len; i++)
{
	if (Attribs[i] != LastAttrib) 
	{
		Tempstr=CatStr(Tempstr,NORM);
		if (Attribs[i] > 0) Tempstr=CatStr(Tempstr,VtCode(Attribs[i] & 0x00FF, Attribs[i] & 0xFF00, Attribs[i] & 0xFF0000));
		LastAttrib=Attribs[i];
	}
	if (LastAttrib & FLAG_CAPS) Tempstr=AddCharToStr(Tempstr,toupper(Line[i]));
	else if (LastAttrib & FLAG_LOWERCASE) Tempstr=AddCharToStr(Tempstr,tolower(Line[i]));
	else Tempstr=AddCharToStr(Tempstr,Line[i]);
}

Tempstr=CatStr(Tempstr,NORM);
write(1,Tempstr,StrLen(Tempstr));

DestroyString(Tempstr);
}



void ColorLine(char *Line, int Len, ListNode *ColorMatches)
{
ListNode *Curr;
int *Attribs=NULL;
char *p_LineEnd;
int i, LastAttrib=0;
TCrayon *Crayon;

Attribs=(int *) calloc(Len,sizeof(int));
Curr=ListGetNext(ColorMatches);
while (Curr)
{
	Crayon=(TCrayon *) Curr->Item;
	switch (Crayon->Type)
	{
	case CRAYON_LINE:
	case CRAYON_LINEMAPTO:
	if (pmatch(Curr->Tag,Line,Len,NULL,PMATCH_SUBSTR) > 0) ApplyActions(Attribs, Line, Line, Line+Len, Crayon);
	break;

	case CRAYON_VALUE:
	case CRAYON_STRING:
	case CRAYON_MAPTO:
		ColorSubstring(Attribs, Line, Len, Crayon);
	break;
	}


	Curr=ListGetNext(Curr);
}

OutputLineWithAttributes(Line, Attribs, Len);

free(Attribs);
}


int HandleStdin(STREAM *StdIn, STREAM *Out)
{
int result;
char *Tempstr=NULL;

Tempstr=SetStrLen(Tempstr,4096);
result=STREAMReadBytes(StdIn,Tempstr,4096);
if (result > 0) 
{
	STREAMWriteBytes(Out,Tempstr,result);
	STREAMFlush(Out);
}

DestroyString(Tempstr);
return(result);
}


int ColorProgramOutput(STREAM *S, ListNode *ColorMatches)
{
	char *Tempstr=NULL, *sptr, *eptr;
	int result, len;

	Tempstr=SetStrLen(Tempstr,4098);
	result=STREAMReadBytes(S,Tempstr,4096);
	if (result > 0)
	{
	Tempstr[result]='\0';

	sptr=Tempstr;
	while (sptr && (*sptr != '\0'))
	{
		eptr=strchr(sptr,'\n');
		if (! eptr) eptr=Tempstr+result;
		else eptr++;
		len=eptr-sptr;
		if (len >0) ColorLine(sptr,len,ColorMatches);
		sptr=eptr;
	}
	}

	DestroyString(Tempstr);
	return(result);
}



void ProcessAppends(ListNode *Crayons, int Type)
{
TCrayon *Item;
ListNode *Curr;
char *Tempstr=NULL;
int *Attribs=NULL;
int i, Len;

Curr=ListGetNext(Crayons);
while (Curr)
{
	Item=(TCrayon *) Curr->Item;
	if (Item->Type==Type)
	{
		Tempstr=SubstituteVarsInString(Tempstr,Item->Match,Vars,0);
		Tempstr=CatStr(Tempstr,"\n");
		Len=StrLen(Tempstr);
		Attribs=realloc(Attribs,Len * sizeof(int));
		memset(Attribs,0,Len*sizeof(int));
		ApplyActions(Attribs, Tempstr, Tempstr, Tempstr+Len,Item); 
		OutputLineWithAttributes(Tempstr, Attribs, Len);
	}
Curr=ListGetNext(Curr);
}

DestroyString(Tempstr);
if (Attribs) free(Attribs);
}	



void ProcessCmdLine(char *CmdLine, ListNode *Crayons)
{
TCrayon *Item;
ListNode *Curr;

Curr=ListGetNext(Crayons);
while (Curr)
{
	Item=(TCrayon *) Curr->Item;
	if (Item->Type==CRAYON_CMDLINE)
	{
		 ColorSubstring(NULL, CmdLine, StrLen(CmdLine), Item);
	}
	Curr=ListGetNext(Curr);
}

}	





//Tempstr=RebuildCommandLine(Tempstr,UserOptions,argc, argv);
STREAM *LaunchCommands(int argc, char *argv[], ListNode *Matches)
{
char *Tempstr=NULL, *Quoted=NULL;
ListNode *Curr;
STREAM *Pipe;
TCrayon *Crayon;
int i;

Curr=ListGetNext(Matches);
while (Curr)
{
Crayon=(TCrayon *) Curr->Item;

if (Crayon->Type==CRAYON_EXEC) 
{
	if (StrLen(Tempstr)) Tempstr=MCatStr(Tempstr,";",Crayon->Match,NULL);
	else Tempstr=CopyStr(Tempstr,Crayon->Match);
}

Curr=ListGetNext(Curr);
}


if (! StrLen(Tempstr)) 
{
	for (i=0; i < argc; i++)
	{
		Quoted=QuoteCharsInStr(Quoted,argv[i]," 	");
		Tempstr=MCatStr(Tempstr,Quoted," ",NULL);
	}
}
Pipe=STREAMSpawnCommand(Tempstr,COMMS_BY_PTY|SPAWN_TRUST_COMMAND);

DestroyString(Tempstr);
DestroyString(Quoted);
return(Pipe);
}


//Currently only reads Title
void XTermReadValue(STREAM *StdIn)
{
char *Tempstr=NULL, *ptr;

Tempstr=CopyStr(Tempstr,"\x1b[21;t");
write(1,Tempstr,StrLen(Tempstr));
Tempstr=STREAMReadToTerminator(Tempstr,StdIn,'\\');
if (StrLen(Tempstr))
{
	ptr=Tempstr+StrLen(Tempstr)-2;
	if (strcmp(ptr,"\x1b\\")==0) *ptr='\0';
	ptr=Tempstr;
	if (strncmp(ptr,"\x1b]l",3)==0) ptr+=3;
	if (StrLen(ptr)) SetVar(Vars,"OldXtermTitle",ptr);
}

DestroyString(Tempstr);
}



//Spawns a command, reads text from it and 'Crayonizes' it
void CrayonizeCommand(int argc, char *argv[])
{
STREAM *StdIn=NULL, *Pipe=NULL, *S;
char *UserOptions=NULL, *CmdLine=NULL, *Tempstr=NULL, *CrayonizerDir=NULL;
ListNode *ColorMatches, *Streams;
int val, i, result;

ColorMatches=ListCreate();
for (i=0; i < argc; i++) CmdLine=MCatStr(CmdLine,argv[i]," ",NULL);
StripTrailingWhitespace(CmdLine);

Tempstr=MCopyStr(Tempstr,GetCurrUserHomeDir(),"/",USER_CONFIG_FILE,NULL);
if (! ConfigReadFile(Tempstr, CmdLine, &UserOptions, &CrayonizerDir, ColorMatches))
{
	if (! ConfigReadFile(GLOBAL_CONFIG_FILE, CmdLine, &UserOptions, &CrayonizerDir, ColorMatches))
	{
		printf("ERROR! Crayonizer can't find config file. Tried %s and %s\n",Tempstr, GLOBAL_CONFIG_FILE);
		exit(1);
	}
}

if (! StrLen(CrayonizerDir))
{
		printf("ERROR! You must specify 'CrayonizerDir' in Crayonizer Config File\n");
		printf("Failure to do so can result in crayonizer calling itself resulting in\n");
		printf("an infinite loop of crayonizer processes (forkbomb)\n");
		printf("AND YOU DON'T WANT THAT!\n");
		exit(1);
}

Tempstr=RebuildPath(Tempstr,getenv("PATH"),CrayonizerDir);
setenv("PATH",Tempstr,TRUE);
LoadEnvironment();

//if not outputing to a tty then run
if (! isatty(1)) 
{
	Tempstr=RebuildCommandLine(Tempstr,UserOptions,argc, argv);
	exit(system(Tempstr));
}

if (! ListSize(ColorMatches)) 
{
	Tempstr=RebuildCommandLine(Tempstr,UserOptions,argc, argv);
	exit(system(Tempstr));
}


//Okay, we're committed to running the command!
InitTTY(0, 0,TTYFLAG_LFCR);
Streams=ListCreate();
StdIn=STREAMFromFD(0);
ListAddItem(Streams,StdIn);



//if we are going to restore the title at the end, then we need to set it here
if (GlobalFlags & FLAG_RESTORE_XTITLE) XTermReadValue(StdIn);


ProcessCmdLine(CmdLine, ColorMatches);
ProcessAppends(ColorMatches,CRAYON_PREPEND);

Pipe=LaunchCommands(argc, argv, ColorMatches);
ListAddItem(Streams,Pipe);

while (1)
{
	S=STREAMSelect(Streams);
	if (S==StdIn) result=HandleStdin(StdIn,Pipe);
	else result=ColorProgramOutput(Pipe, ColorMatches);
	if (result < 0) break;
}
wait(&val);

ProcessAppends(ColorMatches,CRAYON_APPEND);

STREAMClose(Pipe);
STREAMDisassociateFromFD(StdIn);
ResetTTY(0);
ListDestroy(Streams,NULL);
ListDestroy(ColorMatches,free);
DestroyString(Tempstr);
DestroyString(UserOptions);
DestroyString(CrayonizerDir);
}



void CrayonizeSTDIN(int argc, char *argv[])
{
STREAM *StdIn=NULL;
char *UserOptions=NULL, *CmdLine=NULL, *Tempstr=NULL, *CrayonizerDir=NULL;
ListNode *ColorMatches;
int val, i;

ColorMatches=ListCreate();
for (i=2; i < argc; i++) CmdLine=MCatStr(CmdLine,argv[i]," ",NULL);
StripTrailingWhitespace(CmdLine);

Tempstr=MCopyStr(Tempstr,GetCurrUserHomeDir(),"/",USER_CONFIG_FILE,NULL);
if (! ConfigReadFile(Tempstr, CmdLine, &UserOptions, &CrayonizerDir, ColorMatches))
{
	if (! ConfigReadFile(GLOBAL_CONFIG_FILE, CmdLine, &UserOptions, &CrayonizerDir, ColorMatches))
	{
		printf("ERROR! Crayonizer can't find config file. Tried %s and %s\n",Tempstr, GLOBAL_CONFIG_FILE);
		exit(1);
	}
}

Tempstr=RebuildCommandLine(Tempstr,UserOptions,argc, argv);
StdIn=STREAMFromFD(0);

Tempstr=STREAMReadLine(Tempstr,StdIn);
while (Tempstr)
{
ColorLine(Tempstr,StrLen(Tempstr),ColorMatches);
Tempstr=STREAMReadLine(Tempstr,StdIn);
}

STREAMDisassociateFromFD(StdIn);
ListDestroy(ColorMatches,free);
DestroyString(Tempstr);
DestroyString(UserOptions);
DestroyString(CrayonizerDir);
}



void CalledAsSelf(int argc, char *argv[])
{
char *Args[]={"-v","-version","--version","-?","-h","-help","--help","-config-help","-pmatch-help","-stdin",NULL};
typedef enum {ARG_V1, ARG_V2, ARG_V3, ARG_H1, ARG_H2, ARG_H3, ARG_H4, ARG_CONFIG_HELP, ARG_PMATCH_HELP, ARG_STDIN};
int i, val, new_argc=0;
char **new_argv=NULL;


for (i=1; i < argc; i++)
{
	val=MatchTokenFromList(argv[i],Args,0);
	
	switch(val)
	{
		case ARG_V1:
		case ARG_V2:
		case ARG_V3:
			printf("crayonizer (Crayonizer) %s\n",Version);
		break;
		
		case ARG_H1:
		case ARG_H2:
		case ARG_H3:
		case ARG_H4:
			PrintUsage();
		break;

		case ARG_CONFIG_HELP:
			PrintConfigFileHelp();
		break;

		case ARG_PMATCH_HELP:
			PrintPMatchHelp();
		break;

		case ARG_STDIN:
			CrayonizeSTDIN(argc, argv);
		break;

		default:
			new_argc++;
			new_argv=(char **) realloc(new_argv,sizeof(char *) * new_argc);
			new_argv[new_argc-1]=argv[i];
		break;
	}
	
}

if (new_argc > 0) CrayonizeCommand(new_argc,new_argv);
}



main(int argc, char *argv[])
{
char *ptr;

if (strcmp(basename(argv[0]),"crayonizer")==0) CalledAsSelf(argc, argv);
else 
{
ptr=basename(argv[0]);
argv[0]=ptr;
CrayonizeCommand(argc,argv);
}
}
